<?php

declare(strict_types=1);

namespace Drupal\file_defaults_by_type\Plugin\Field\FieldWidget;

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\FieldDefinitionInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Drupal\Core\Field\WidgetInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Element;
use Drupal\Core\Render\ElementInfoManagerInterface;
use Drupal\Core\Session\AccountProxyInterface;
use Drupal\file\Element\ManagedFile;
use Drupal\file\FileRepositoryInterface;
use Drupal\file\Plugin\Field\FieldWidget\FileWidget;
use Drupal\file_defaults_by_type\FileLocationTrait;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the Default file choice widget.
 *
 * @FieldWidget(
 *   id = "default_file_choice",
 *   label = @Translation("Default file choice"),
 *   field_types = {
 *     "file"
 *   }
 * )
 */
class DefaultFileChoiceWidget extends FileWidget implements WidgetInterface {

  use FileLocationTrait;

  /**
   * {@inheritdoc}
   */
  public function __construct(
    $plugin_id,
    $plugin_definition,
    FieldDefinitionInterface $field_definition,
    array $settings,
    array $third_party_settings,
    ElementInfoManagerInterface $element_info,
    protected EntityRepositoryInterface $entityRepository,
    protected EntityTypeManagerInterface $entityTypeManager,
    protected FileRepositoryInterface $fileRepository,
    protected AccountProxyInterface $currentUser,
  ) {
    parent::__construct(
      $plugin_id,
      $plugin_definition,
      $field_definition,
      $settings,
      $third_party_settings,
      $element_info
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['third_party_settings'],
      $container->get('element_info'),
      $container->get('entity.repository'),
      $container->get('entity_type.manager'),
      $container->get('file.repository'),
      $container->get('current_user'),
    );
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state): array {
    $element = parent::formElement($items, $delta, $element, $form, $form_state);
    $settings = $this->fieldDefinition->getThirdPartySettings('file_defaults_by_type');

    // Once uploaded we cannot change type.
    $default_fids = $element['#extended'] ? $element['#default_value']['fids'] : $element['#default_value'];
    if (!empty($default_fids)) {
      return $element;
    }

    // Get type from state.
    $type = $form_state->getValue([$element['#field_name'], $delta, 'default_file_choice', 'file_type']);

    // Minimum required is label and default.
    $options = [];
    $file_name = '';
    $settings = $this->fieldDefinition->getThirdPartySettings('file_defaults_by_type');
    foreach ($settings['types'] as $option) {
      if (!empty($option['label']) && !empty($option['default'])) {
        /** @var \Drupal\file\FileInterface $file */
        $file = $this->entityRepository->loadEntityByUuid('file', $option['default']);
        if ($file !== NULL) {
          // Use fids for later use.
          $options[$file->id()] = $option['label'];
          // We store file name for description, it will be first element or
          // type from state.
          if ($file_name === '' || (isset($type) && $file->id() == $type)) {
            $file_name = $file->getFilename();
          }
        }
      }
    }

    if (empty($options)) {
      return $element;
    }

    $element['default_file_choice'] = [
      '#type' => 'container',
    ];

    // Type select.
    $element['default_file_choice']['file_type'] = [
      '#type' => 'select',
      '#title' => $this->t('File type'),
      '#default_value' => $options[$type] ?? array_key_first($options),
      '#options' => $options,
      '#description' => $this->t(
        'Select the desired file type. Click on <i>Attach</i> to set default file: <strong>@filename</strong>.',
        ['@filename' => $file_name]
      ),
      '#ajax' => [
        'callback' => [$this, 'updateAjaxCallback'],
        // Wrapper is set in $this->process().
      ],
    ];

    // Button to attach default files.
    // We need to use 'upload_button' as key for the submit, to pass
    // \Drupal\file\Element\ManagedFile:validateManagedFile() validation.
    // Any required field needs at least to have fid values or use that key.
    $element['default_file_choice']['upload_button'] = [
      '#type' => 'submit',
      '#submit' => [[$this, 'attach']],
      '#value' => $this->t('Attach'),
      '#ajax' => [
        'callback' => [ManagedFile::class, 'uploadAjaxCallback'],
        // Wrapper is set in $this->process().
      ],
    ];

    return $element;
  }

  /**
   * {@inheritdoc}
   *
   * Adds ajax wrapper and options for choice children.
   */
  public static function process($element, FormStateInterface $form_state, $form): array {
    $element = parent::process($element, $form_state, $form);

    if (!isset($element['default_file_choice'])) {
      return $element;
    }

    // Get wrapper and options from processed element.
    // Upload button has, at this point, a target wrapper and query options.
    $wrapper = '';
    $options = [];
    foreach (Element::children($element) as $key) {
      if (isset($element[$key]['#ajax'])) {
        $wrapper = $element[$key]['#ajax']['wrapper'];
        $options = $element[$key]['#ajax']['options'];
        break;
      }
    }

    // Propagate Ajax configuration to the choice element children.
    foreach (Element::children($element['default_file_choice']) as $key) {
      if (isset($element['default_file_choice'][$key]['#ajax'])) {
        $element['default_file_choice'][$key]['#ajax']['wrapper'] = $wrapper;
        $element['default_file_choice'][$key]['#ajax']['options'] = $options;
      }
    }

    // Hide upload options from parent widget.
    $element['upload']['#access'] = FALSE;
    $element['upload_button']['#access'] = FALSE;
    $element['remove_button']['#access'] = FALSE;

    $element['default_file_choice']['upload_button']['#limit_validation_errors'] = [$element['#parents']];

    return $element;
  }

  /**
   * Submit handler to attach the default file.
   *
   * @param mixed $form
   *   The build form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function attach(array $form, FormStateInterface $form_state): void {
    $button = $form_state->getTriggeringElement();

    // Get value from select: button parent (default_file_choice) > select type.
    $type = $form_state->getValue(array_slice($button['#parents'], 0, -1))['file_type'];
    if ($type === NULL) {
      return;
    }

    /** @var \Drupal\file\FileInterface $file */
    $file = $this->entityTypeManager->getStorage('file')->load($type);
    if ($file === NULL) {
      return;
    }

    // Copy file using destination from field settings.
    $settings = $this->fieldDefinition->getItemDefinition()->getSettings();
    $upload_location = $this->getUploadLocation($settings);
    /** @var \Drupal\file\FileInterface $copy */
    $copy = $this->fileRepository->copy($file, $upload_location . '/' . $file->getFilename());
    $copy->setTemporary();
    // Temporary files have to be owned by the current user.
    // @see \Drupal\file\Element\ManagedFile:valueCallback()
    // @todo handle anonymous file ownership.
    $copy->setOwnerId($this->currentUser->id());
    $copy->save();

    // Set values in the state. These values are discarded since the validation
    // checks that they are present in the request. This prevents this widget to
    // display the select and attach button. Just add new value to the existing
    // items.
    // @see file_managed_file_save_upload()
    $element = NestedArray::getValue($form, array_slice($button['#array_parents'], 0, -2));
    $field_name = $element['#field_name'];
    $parents = $element['#field_parents'];
    $field_state = static::getWidgetState($parents, $field_name, $form_state);
    $field_state['items'][] = $copy->id();
    static::setWidgetState($parents, $field_name, $form_state, $field_state);

    // Set the file like if the user has submitted.
    // @see \Drupal\file\Element\ManagedFile::valueCallback();
    $user_input = $form_state->getUserInput();
    $parents = array_slice($button['#parents'], 0, -2);
    $parents[] = 'fids';
    NestedArray::setValue($user_input, $parents, $copy->id());
    $form_state->setUserInput($user_input);

    $form_state->setRebuild();
  }

  /**
   * Ajax callback for widget replacement.
   *
   * @param mixed $form
   *   The build form.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   *
   * @return array
   *   The processed element.
   */
  public function updateAjaxCallback(array $form, FormStateInterface $form_state): array {
    $trigger = $form_state->getTriggeringElement();
    // Go to widget level: widget > delta > select/attach.
    $parents = array_slice($trigger['#array_parents'], 0, -3);
    $element = NestedArray::getValue($form, $parents);

    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public static function isApplicable(FieldDefinitionInterface $field_definition) {
    $settings = $field_definition->getThirdPartySettings('file_defaults_by_type');

    return !empty($settings['enabled']) && !empty($settings['types']);
  }

}
