<?php

declare(strict_types=1);

namespace Drupal\file_defaults_by_type;

use Drupal\Component\Utility\Bytes;
use Drupal\Component\Utility\Environment;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityRepositoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\file\Element\ManagedFile;
use Drupal\multivalue_form_element\Element\MultiValue;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Class to manage file field settings.
 *
 * @internal
 */
final class SettingsFormAlter implements ContainerInjectionInterface {

  use StringTranslationTrait;
  use FileLocationTrait;

  /**
   * Creates a new instance of this class.
   *
   * @param \Drupal\Core\Entity\EntityRepositoryInterface $entityRepository
   *   The entity type manager.
   * @param \Drupal\Core\Render\RendererInterface $renderer
   *   The renderer.
   */
  public function __construct(
    protected EntityRepositoryInterface $entityRepository,
    protected RendererInterface $renderer,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): static {
    return new static(
      $container->get('entity.repository'),
      $container->get('renderer'),
    );
  }

  /**
   * Alters the form to add types settings.
   *
   * @param array $form
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public function settingsFormAlter(array &$form, FormStateInterface $form_state): void {
    $storage = $form_state->getStorage();

    if ($storage === NULL || !isset($storage['field_config']) || $storage['field_config']->get('field_type') !== 'file') {
      return;
    }

    /** @var \Drupal\field\Entity\FieldConfig $field_config */
    $field_config = $storage['field_config'];
    $default_settings = $field_config->getThirdPartySettings('file_defaults_by_type');

    $settings['file_defaults_by_type']['enabled'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Default file types enabled'),
      '#default_value' => $enabled = $default_settings['enabled'] ?? FALSE,
      '#description' => $this->t('Enable file types: label and default file.'),
    ];

    // Prepare default values.
    $default_types = $enabled ? $default_settings['types'] : [];
    foreach ($default_types as &$default_type) {
      // Replace UUID with fids in default file.
      // Values are converted into UUIDs later on validation.
      if (!empty($default_type['default'])) {
        $file = $this->entityRepository->loadEntityByUuid('file', $default_type['default']);
        $default_type['default'] = $file ? [$file->id()] : [];
      }
    }

    // @todo Label should be required in all types.
    $settings['file_defaults_by_type']['types'] = [
      '#type' => 'multivalue',
      '#title' => $this->t('Types'),
      '#cardinality' => MultiValue::CARDINALITY_UNLIMITED,
      '#default_value' => $default_types,
      '#states' => [
        'visible' => [
          ':input[name="third_party_settings[file_defaults_by_type][enabled]"]' => ['checked' => TRUE],
        ],
      ],
    ];

    $settings['file_defaults_by_type']['types']['label'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Label'),
      '#description' => $this->t("Label displayed to the user to select the type used."),
    ];

    // Get upload validators based on field settings.
    $field_settings = $field_config->getSettings();
    // @todo Use Drupal\file\Validation\FileValidatorSettingsTrait when support
    // for Drupal Core 10.2.x is dropped.
    $upload_validators = $this->getFileUploadValidators($field_settings);

    $settings['file_defaults_by_type']['types']['default'] = [
      '#type' => 'managed_file',
      '#title' => $this->t('Default file'),
      '#field_name' => $field_config->getName(),
      '#upload_location' => $this->getUploadLocation($field_settings),
      '#element_validate' => [
        [ManagedFile::class, 'validateManagedFile'],
        [static::class, 'validateDefaultFileForm'],
      ],
      // @todo Update allowed extensions based on field allowed extensions
      // settings. Handle interaction between uploaded default values and
      // allowed file extensions. Eg: A change in extensions field could make
      // some of the fields invalid, all need to be validated again.
      '#upload_validators' => $upload_validators,
      '#description' => [
        '#theme' => 'file_upload_help',
        '#description' => 'File to be added if no file is uploaded.',
        '#upload_validators' => $upload_validators,
        '#cardinality' => 1,
      ],
    ];

    // Make default files permanent.
    $form['actions']['submit']['#submit'][] = [static::class, 'typeDefaultFilesPermanent'];

    // Add settings to the form.
    $form['third_party_settings'] += $settings;
  }

  /**
   * Submit handler to make uploaded files for default types permanent.
   *
   * @param array $form
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public static function typeDefaultFilesPermanent(array &$form, FormStateInterface $form_state): void {
    $settings = $form_state->getValue('third_party_settings');

    if (empty($settings['file_defaults_by_type']['types'])) {
      return;
    }

    foreach ($settings['file_defaults_by_type']['types'] as $type) {
      if (!isset($type['default']) || empty($type['default'])) {
        continue;
      }

      /** @var \Drupal\file\Entity\File $file */
      $file = \Drupal::service('entity.repository')->loadEntityByUuid('file', $type['default']);

      // File doesn't exist or is already a permanent file.
      if ($file === NULL || $file->isPermanent()) {
        continue;
      }

      // Make the submitted file permanent.
      $file->setPermanent();
      $file->save();
    }
  }

  /**
   * Validates the managed_file element for the default File form.
   *
   * Converts fid into UUID.
   *
   * @param array $element
   *   The form element to process.
   * @param \Drupal\Core\Form\FormStateInterface $form_state
   *   The form state.
   */
  public static function validateDefaultFileForm(array &$element, FormStateInterface $form_state): void {
    $value = '';

    if (isset($element['fids']['#value'][0])) {
      /** @var \Drupal\file\Entity\File $file */
      if ($file = \Drupal::entityTypeManager()->getStorage('file')->load($element['fids']['#value'][0])) {
        $value = $file->uuid();
      }
    }

    $form_state->setValueForElement($element, $value);
  }

  /**
   * Retrieves the upload validators for a file field.
   *
   * @param array $settings
   *   The field settings.
   *
   * @return array
   *   An array suitable for passing to file_save_upload() or the file field
   *   element's '#upload_validators' property.
   */
  public function getFileUploadValidators(array $settings): array {
    $validators = [];

    // Cap the upload size according to the PHP limit.
    $max_filesize = Bytes::toNumber(Environment::getUploadMaxSize());
    if (!empty($settings['max_filesize'])) {
      $max_filesize = min($max_filesize, Bytes::toNumber($settings['max_filesize']));
    }

    // There is always a file size limit due to the PHP server limit.
    $validators['FileSizeLimit'] = ['fileLimit' => $max_filesize];

    // Add the extension check if necessary.
    if (!empty($settings['file_extensions'])) {
      $validators['FileExtension'] = ['extensions' => $settings['file_extensions']];
    }

    return $validators;
  }

}
