<?php

declare(strict_types=1);

namespace Drupal\Tests\file_defaults_by_type\FunctionalJavascript;

use Drupal\FunctionalJavascriptTests\WebDriverTestBase;
use Drupal\Tests\TestFileCreationTrait;
use Drupal\file\Entity\File;

/**
 * Tests the file types settings.
 */
class FieldSettingsFormTest extends WebDriverTestBase {

  use TestFileCreationTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'node',
    'file_defaults_by_type',
    'field_ui',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * Test settings added for File field form.
   */
  public function testSettingsForm(): void {
    $this->drupalCreateContentType(['type' => 'article', 'name' => 'Article']);
    $adminUser = $this->drupalCreateUser([
      'access administration pages',
      'administer content types',
      'administer node fields',
    ]);
    $this->drupalLogin($adminUser);
    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();
    $test_files = $this->getTestFiles('text');

    // Add field using 'File upload' type to check changes on field form.
    $this->drupalGet('admin/structure/types/manage/article/fields/add-field');
    $page->find('xpath', '//*[text() = "File upload"]')->getParent()->click();
    $page->pressButton('Continue');
    $page->fillField('label', 'File test field');
    $page->find('xpath', '//*[text() = "File"]')->getParent()->click();
    $page->pressButton('Continue');

    // Fields are present.
    $enabled_field = $assert_session->fieldExists('Default file types enabled');
    $types_selector = '[data-drupal-selector="edit-third-party-settings-file-defaults-by-type-types"]';
    $assert_session->elementTextNotContains('css', $types_selector, 'Types');

    // After checking enabled field: types are displayed.
    $enabled_field->check();
    $assert_session->elementTextContains('css', $types_selector, 'Types');

    // Names helper functions.
    $field_name = fn (string $field, string $delta): string => "third_party_settings[file_defaults_by_type][types][$delta][$field]";
    $upload_name = fn (string $delta): string => "third_party_settings_file_defaults_by_type_types_{$delta}_default";

    // Add different types and save.
    // Complete type: label and default.
    $text_file_1 = $test_files[0];
    $test_file_path = \Drupal::service('file_system')->realpath($text_file_1->uri);
    $types = $assert_session->elementExists('css', $types_selector);
    $assert_session->fieldExists($field_name('label', '0'), $types)->setValue('Txt');
    $page->attachFileToField("files[{$upload_name('0')}]", $test_file_path);
    $assert_session->waitForElementVisible('css', "[name='files[{$upload_name('0')}_remove_button']");
    // File is created but not permanent.
    $file_1_id = $assert_session->elementExists('css', "[name='{$field_name('default', '0')}[fids]']")->getValue();
    $file_1 = File::load($file_1_id);
    $this->assertFalse($file_1->isPermanent());

    // Add another item, only label.
    $assert_session->buttonExists('Add another item')->press();
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->fieldExists($field_name('label', '1'), $types)->setValue('Word');

    // Add another item, only default.
    $assert_session->buttonExists('Add another item')->press();
    $assert_session->assertWaitOnAjaxRequest();
    $text_file_2 = $test_files[1];
    $test_file_path = \Drupal::service('file_system')->realpath($text_file_2->uri);
    $page->attachFileToField("files[{$upload_name('2')}]", $test_file_path);
    $assert_session->waitForElementVisible('css', "[name='files[{$upload_name(delta: '2')}_remove_button']");
    // File is created but not permanent.
    $file_2_id = $assert_session->elementExists('css', "[name='{$field_name('default', '2')}[fids]']")->getValue();
    $file_2 = File::load($file_2_id);
    $this->assertFalse($file_2->isPermanent());

    $assert_session->buttonExists('Save settings')->press();
    $assert_session->statusMessageContains('Saved File test field configuration.', 'status');

    // Check values: files.
    // Check that submitted files are now permanent.
    \Drupal::entityTypeManager()->getStorage(entity_type_id: 'file')->resetCache();
    $file_1 = File::load($file_1_id);
    $this->assertTrue($file_1->isPermanent());
    $file_2 = File::load($file_2_id);
    $this->assertTrue($file_2->isPermanent());

    // Check values: form.
    $this->drupalGet('admin/structure/types/manage/article/fields/node.article.field_file_test_field');
    // Enabled types.
    $this->assertTrue($assert_session->fieldExists('Default file types enabled')->isChecked());
    $types = $assert_session->elementExists('css', $types_selector);
    // Full type.
    $this->assertEquals('Txt', $assert_session->fieldExists($field_name('label', '0'), $types)->getValue());
    $assert_session->elementTextContains('css', '#edit-third-party-settings-file-defaults-by-type-types-0-default', $text_file_1->name);
    // Only label.
    $this->assertEquals('Word', $assert_session->fieldExists($field_name('label', '1'), $types)->getValue());
    // Only file.
    $assert_session->elementTextContains('css', '#edit-third-party-settings-file-defaults-by-type-types-2-default', $text_file_2->name);

    // Disable types and save, then check that values are removed.
    $this->drupalGet('admin/structure/types/manage/article/fields/node.article.field_file_test_field');
    $assert_session->fieldExists(field: 'Default file types enabled')->uncheck();
    $assert_session->buttonExists('Save settings')->press();
    $assert_session->statusMessageContains('Saved File test field configuration.', 'status');
    // Check that configuration is removed.
    $this->drupalGet('admin/structure/types/manage/article/fields/node.article.field_file_test_field');
    $assert_session->fieldExists('Default file types enabled')->check();
    $types = $assert_session->elementExists('css', $types_selector);
    $this->assertEquals('', $assert_session->fieldExists($field_name('label', '0'), $types)->getValue());
    $assert_session->elementExists('css', "[name='files[{$upload_name('0')}]']", $types);

    // Types default files are restricted to field allowed extensions.
    $this->drupalGet('admin/structure/types/manage/article/fields/node.article.field_file_test_field');
    $assert_session->fieldExists('Allowed file extensions')->setValue('test');
    $assert_session->buttonExists('Save settings')->press();
    $assert_session->statusMessageContains('Saved File test field configuration.', 'status');
    // The user cannot upload files with different extension.
    $this->drupalGet('admin/structure/types/manage/article/fields/node.article.field_file_test_field');
    $assert_session->fieldExists(field: 'Default file types enabled')->check();
    $page->attachFileToField("files[{$upload_name('0')}]", $test_file_path);
    $messages = $assert_session->waitForElementVisible('css', '.file-upload-js-error');
    $this->assertEquals("The selected file {$test_files[1]->name}.txt cannot be uploaded. Only files with the following extensions are allowed: test.", $messages->getText());
  }

}
