<?php

declare(strict_types=1);

namespace Drupal\Tests\file_defaults_by_type\FunctionalJavascript;

use Drupal\FunctionalJavascriptTests\WebDriverTestBase;
use Drupal\Tests\file\Functional\FileFieldCreationTrait;
use Drupal\Tests\file_defaults_by_type\Traits\TestsFilesTrait;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;

/**
 * Tests the file field widget, single and multi-valued, using AJAX upload.
 */
class WidgetUploadTest extends WebDriverTestBase {

  use FileFieldCreationTrait;
  use TestsFilesTrait;

  /**
   * An user with administration permissions.
   *
   * @var \Drupal\user\UserInterface
   */
  protected $adminUser;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'node',
    'file_defaults_by_type',
  ];

  /**
   * {@inheritdoc}
   */
  protected $defaultTheme = 'stark';

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->drupalCreateContentType(['type' => 'article', 'name' => 'Article']);
    $this->adminUser = $this->drupalCreateUser([
      'access content',
      'access administration pages',
      'administer nodes',
      'bypass node access',
    ]);
    $this->drupalLogin($this->adminUser);

    // Create file field with some types.
    $this->createFileField(
      'test_file_field',
      'node',
      'article',
      [
        'cardinality' => 1,
      ],
      [
        'file_directory' => '',
      ]
    );

    $this->generateTestFiles();
    $settings = [
      'enabled' => TRUE,
      'types' => [
        [
          'label' => "File 1",
          'default' => "uuid-1",
        ],
        [
          'label' => "File 2",
          'default' => "uuid-2",
        ],
        [
          'label' => "File 3",
          'default' => "uuid-3",
        ],
      ],
    ];

    foreach ($settings as $k => $v) {
      FieldConfig::load('node.article.test_file_field')
        ->setThirdPartySetting('file_defaults_by_type', $k, $v)->save();
    }

    // Change widget to use the default files.
    \Drupal::service('entity_display.repository')->getFormDisplay('node', 'article')
      ->setComponent('test_file_field', ['type' => 'default_file_choice'])
      ->save();
  }

  /**
   * Tests default_file_choice for single valued widget.
   */
  public function testSingleValuedWidget(): void {
    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    $this->drupalGet('node/add/article');
    $title = $this->randomString();
    $page->findField('title[0][value]')->setValue($title);
    // Check that existing files are only generated.
    $this->assertFilesCount(3);
    $uid = (int) \Drupal::currentUser()->id();

    // Check that value is set using attach.
    $assert_session->selectExists('File type')->selectOption('File 2');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->buttonExists('Attach')->press();
    $assert_session->waitForElementVisible('css', '[name="test_file_field_0_remove_button"]');
    $assert_session->hiddenFieldValueEquals('test_file_field[0][fids]', '4');
    $this->assertFilesCount(4);
    $this->assertCopyFile(2, 4, $uid);
    $page->pressButton('Save');
    $assert_session->statusMessageContains("Article $title has been created.", 'status');
    $assert_session->linkExists('example-2.txt');

    // Check that the value can be replaced.
    $this->drupalGet('node/1/edit');
    $assert_session->elementExists('css', '[name="test_file_field_0_remove_button"]')->press();
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->selectExists('File type')->selectOption('File 3');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->buttonExists('Attach')->press();
    $assert_session->waitForElementVisible('css', '[name="test_file_field_0_remove_button"]');
    $assert_session->hiddenFieldValueEquals('test_file_field[0][fids]', '5');
    $this->assertFilesCount(5);
    $this->assertCopyFile(3, 5, $uid);
    $page->pressButton('Save');
    $assert_session->statusMessageContains("Article $title has been updated.", 'status');
    $assert_session->linkExists('example-3.txt');
    $assert_session->linkNotExists('example-2.txt');

    // Check that we can change the type, but we can opt to leave it empty.
    $this->drupalGet('node/1/edit');
    $assert_session->elementExists('css', '[name="test_file_field_0_remove_button"]')->press();
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->selectExists('File type')->selectOption('File 2');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->selectExists('File type')->selectOption('File 1');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->selectExists('File type')->selectOption('File 3');
    $assert_session->assertWaitOnAjaxRequest();
    $this->assertFilesCount(5);
    $page->pressButton('Save');
    $assert_session->statusMessageContains("Article $title has been updated.", 'status');
    $assert_session->linkNotExists('example-1.txt');
    $assert_session->linkNotExists('example-2.txt');
    $assert_session->linkNotExists('example-3.txt');

    // Check required field, try with an empty value, set value and save.
    FieldConfig::load('node.article.test_file_field')->setRequired(TRUE)->save();
    $this->drupalGet('node/1/edit');
    $assert_session->selectExists('File type')->selectOption('File 2');
    $assert_session->assertWaitOnAjaxRequest();
    $page->pressButton('Save');
    $assert_session->statusMessageContains('test_file_field field is required.', 'error');
    $assert_session->buttonExists('Attach')->press();
    $assert_session->waitForElementVisible('css', '[name="test_file_field_0_remove_button"]');
    $assert_session->hiddenFieldValueEquals('test_file_field[0][fids]', '6');
    $this->assertFilesCount(6);
    $this->assertCopyFile(2, 6, $uid);
    $page->pressButton('Save');
    $assert_session->statusMessageContains("Article $title has been updated.", 'status');
    $assert_session->linkExists('example-2.txt');
  }

  /**
   * Tests default_file_choice for multivalued valued widget.
   */
  public function testMultiValuedWidget(): void {
    FieldStorageConfig::load('node.test_file_field')->setCardinality(-1)->save();
    $assert_session = $this->assertSession();
    $page = $this->getSession()->getPage();

    $this->drupalGet('node/add/article');
    $title = $this->randomString();
    $page->findField('title[0][value]')->setValue($title);
    $this->assertFilesCount(3);
    $uid = (int) \Drupal::currentUser()->id();

    // Check that multiple values are set using attach.
    $assert_session->buttonExists('Attach')->press();
    $assert_session->waitForElementVisible('css', '[name="test_file_field_0_remove_button"]');
    $assert_session->selectExists('File type')->selectOption('File 2');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->buttonExists('Attach')->press();
    $assert_session->waitForElementVisible('css', '[name="test_file_field_1_remove_button"]');
    $assert_session->selectExists('File type')->selectOption('File 3');
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->buttonExists('Attach')->press();
    $assert_session->waitForElementVisible('css', '[name="test_file_field_2_remove_button"]');
    $assert_session->hiddenFieldValueEquals('test_file_field[0][fids]', '4');
    $assert_session->hiddenFieldValueEquals('test_file_field[1][fids]', '5');
    $assert_session->hiddenFieldValueEquals('test_file_field[2][fids]', '6');
    $this->assertFilesCount(6);
    $this->assertCopyFile(1, 4, $uid);
    $this->assertCopyFile(2, 5, $uid);
    $this->assertCopyFile(3, 6, $uid);
    $page->pressButton('Save');
    $assert_session->statusMessageContains("Article $title has been created.", 'status');
    $assert_session->linkExists('example-1.txt');
    $assert_session->linkExists('example-2.txt');
    $assert_session->linkExists('example-3.txt');

    // Check the removal or all files, and empty value for the field.
    $this->drupalGet('node/1/edit');
    $assert_session->elementExists('css', '[name="test_file_field_0_remove_button"]')->press();
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->elementExists('css', '[name="test_file_field_0_remove_button"]')->press();
    $assert_session->assertWaitOnAjaxRequest();
    $assert_session->elementExists('css', '[name="test_file_field_0_remove_button"]')->press();
    $assert_session->assertWaitOnAjaxRequest();
    $this->assertFilesCount(6);
    $page->pressButton('Save');
    $assert_session->statusMessageContains("Article $title has been updated.", 'status');
    $assert_session->linkNotExists('example-1.txt');
    $assert_session->linkNotExists('example-2.txt');
    $assert_session->linkNotExists('example-3.txt');

    // Check required field, try with an empty value, set value and save.
    FieldConfig::load('node.article.test_file_field')->setRequired(TRUE)->save();
    $this->drupalGet('node/1/edit');
    $assert_session->selectExists('File type')->selectOption('File 2');
    $assert_session->assertWaitOnAjaxRequest();
    $page->pressButton('Save');
    $assert_session->statusMessageContains('test_file_field field is required.', 'error');
    $assert_session->buttonExists('Attach')->press();
    $assert_session->waitForElementVisible('css', '[name="test_file_field_0_remove_button"]');
    $assert_session->hiddenFieldValueEquals('test_file_field[0][fids]', '7');
    $this->assertCopyFile(2, 7, $uid);
    $this->assertFilesCount(7);
    $page->pressButton('Save');
    $assert_session->statusMessageContains("Article $title has been updated.", 'status');
    $assert_session->linkExists('example-2.txt');
    $assert_session->linkNotExists('example-1.txt');
    $assert_session->linkNotExists('example-3.txt');
  }

  /**
   * Disables the native browser validation for required fields.
   */
  protected function disableNativeBrowserRequiredFieldValidation(): void {
    $this->getSession()->executeScript('jQuery(":input[required]").prop("required", false);');
  }

  /**
   * Asserts the copy version of the file on widget upload.
   *
   * The values are the same except for the file URI, author and status.
   *
   * @param int $origin_fid
   *   The source File.
   * @param int $copy_fid
   *   The copy version of the File.
   * @param int $uid
   *   The user id from the author of the file.
   */
  protected function assertCopyFile(int $origin_fid, int $copy_fid, int $uid): void {
    $file_storage = \Drupal::entityTypeManager()->getStorage('file');

    $this->assertNotEquals($origin_fid, $copy_fid);

    /** @var \Drupal\file\Entity\File $origin */
    $origin = $file_storage->load($origin_fid);
    if ($origin === NULL) {
      throw new \InvalidArgumentException("The origin File with the fid:$origin_fid doesn't exist.");
    }

    /** @var \Drupal\file\Entity\File $copy */
    $copy = $file_storage->load($copy_fid);
    if ($copy === NULL) {
      throw new \InvalidArgumentException("The copied File with the fid:$copy_fid doesn't exist.");
    }

    // Check properties.
    $this->assertTrue($copy->isTemporary());
    $this->assertEquals($uid, $copy->getOwner()->id());
    $this->assertNotEquals($origin->getFileUri(), $copy->getFileUri());
    $this->assertEquals($origin->getFilename(), $copy->getFilename());
    $this->assertEquals($origin->getSize(), $copy->getSize());
    $this->assertEquals($origin->getMimeType(), $copy->getMimeType());
    $this->assertEquals(file_get_contents($origin->getFileUri()), file_get_contents($copy->getFileUri()));
  }

}
