<?php

declare(strict_types=1);

namespace Drupal\Tests\file_defaults_by_type\Kernel;

use Drupal\Component\Utility\Environment;
use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Form\FormState;
use Drupal\Core\StringTranslation\ByteSizeMarkup;
use Drupal\KernelTests\Core\Entity\EntityKernelTestBase;
use Drupal\Tests\file_defaults_by_type\Traits\TestsFilesTrait;
use Drupal\entity_test\Entity\EntityTest;
use Drupal\field\Entity\FieldConfig;
use Drupal\field\Entity\FieldStorageConfig;

/**
 * Tests the rendering of default_file_choice widget field.
 */
class RenderWidgetTest extends EntityKernelTestBase {

  use TestsFilesTrait;

  /**
   * {@inheritdoc}
   */
  protected static $modules = [
    'file',
    'file_defaults_by_type',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp(): void {
    parent::setUp();

    $this->installEntitySchema('file');
    $this->installSchema('file', ['file_usage']);
    $this->installConfig(['file', 'file_defaults_by_type']);

    $this->generateTestFiles();
  }

  /**
   * Tests the File defaults by type single value widget.
   *
   * @dataProvider dataProvider
   */
  public function testFileMultipleWidget(array $expected, array $options): void {
    $form = $this->buildEntityTestForm($options, -1);
    $this->assertWidgetSelect($expected, $form);

    // Check widget description.
    $default_max_size = ByteSizeMarkup::create(Environment::getUploadMaxSize());
    $this->assertEquals(
      "Unlimited number of files can be uploaded to this field.$default_max_size limit.Allowed types: txt.\n",
      $form['test_file_field']['widget']['0']['#description']
    );
  }

  /**
   * Tests the File defaults by type single value widget.
   *
   * @dataProvider dataProvider
   */
  public function testFileSingleWidget(array $expected, array $options): void {
    $form = $this->buildEntityTestForm($options, 1);
    $this->assertWidgetSelect($expected, $form);

    // Check widget description.
    $default_max_size = ByteSizeMarkup::create(Environment::getUploadMaxSize());
    $this->assertEquals(
      "One file only.$default_max_size limit.Allowed types: txt.\n",
      $form['test_file_field']['widget']['0']['#description']
    );
  }

  /**
   * Assert function to check select.
   *
   * @param array $expected
   *   Expected values.
   * @param array $form
   *   The build form.
   */
  protected function assertWidgetSelect(array $expected, array $form): void {
    // Check options in select.
    if (isset($expected['options'])) {
      $this->assertEquals(
        $expected['options'],
        $form['test_file_field']['widget']['0']['default_file_choice']['file_type']['#options']
      );

      // Check button.
      $this->assertEquals(
        "Attach",
        $form['test_file_field']['widget']['0']['default_file_choice']['upload_button']['#value']
      );

      // Select description based on default file.
      if (isset($expected['default'])) {
        $this->assertEquals(
          "Select the desired file type. Click on <i>Attach</i> to set default file: <strong>{$expected['default']}</strong>.",
          $form['test_file_field']['widget']['0']['default_file_choice']['file_type']['#description']
        );
      }
    }
    else {
      $this->assertTrue(!isset($form['test_file_field']['widget']['0']['default_file_choice']));
    }
  }

  /**
   * Data provider for File defaults by type widget.
   *
   * @return \Generator
   *   The test data.
   */
  protected function dataProvider(): \Generator {
    // Disabled type no types: no select.
    yield [
      'expected' => [],
      'settings' => [
        'enabled' => FALSE,
        'types' => [],
      ],
    ];
    // Disabled type and valid type: no select.
    yield [
      'expected' => [],
      'settings' => [
        'enabled' => FALSE,
        'types' => [
          [
            'label' => 'File 1',
            'default' => 'uuid-1',
          ],
        ],
      ],
    ];
    // Enabled type and no types: no select.
    yield [
      'expected' => [],
      'settings' => [
        'enabled' => TRUE,
        'types' => [],
      ],
    ];
    // Enabled type and no valid types: no select.
    yield [
      'expected' => [],
      'settings' => [
        'enabled' => TRUE,
        'types' => [
          [
            'label' => 'File 1',
          ],
          [
            'default' => 'uuid-1',
          ],
        ],
      ],
    ];
    // Enabled type and one valid type: select with one option.
    yield [
      'expected' => [
        'options' => [
          '1' => 'File 1',
        ],
        'default' => 'example-1.txt',
      ],
      'settings' => [
        'enabled' => TRUE,
        'types' => [
          [
            'label' => 'File 1',
            'default' => 'uuid-1',
          ],
        ],
      ],
    ];
    // Enabled multiple types: select with three options.
    yield [
      'expected' => [
        'options' => [
          '1' => 'File 1',
          '2' => 'File 2',
          '3' => 'File 3',
        ],
        'default' => 'example-1.txt',
      ],
      'settings' => [
        'enabled' => TRUE,
        'types' => [
          [
            'label' => 'File 1',
            'default' => 'uuid-1',
          ],
          [
            'label' => 'File 2',
            'default' => 'uuid-2',
          ],
          [
            'label' => 'File 3',
            'default' => 'uuid-3',
          ],
        ],
      ],
    ];
    // Enabled type and one valid label with default wrong file ID.
    yield [
      'expected' => [],
      'settings' => [
        'enabled' => TRUE,
        'types' => [
          [
            'label' => 'File 1',
            'default' => 'wrong-uuid',
          ],
        ],
      ],
    ];
  }

  /**
   * Builds a form using the test entity with File defaults by type widget.
   *
   * @param array $settings
   *   Type settings to be applied.
   * @param int $cardinality
   *   Cardinality for the field.
   *
   * @return array
   *   The built form.
   */
  protected function buildEntityTestForm(array $settings, int $cardinality): array {
    $entity = EntityTest::create();
    FieldStorageConfig::create([
      'field_name' => 'test_file_field',
      'entity_type' => 'entity_test',
      'type' => 'file',
      'cardinality' => $cardinality,
    ])->save();

    $field = FieldConfig::create(values: [
      'label' => 'File field',
      'field_name' => 'test_file_field',
      'entity_type' => 'entity_test',
      'bundle' => 'entity_test',
      'settings' => [
        'file_directory' => $this->getRandomGenerator()->name(8),
      ],
    ]);
    $field->save();

    foreach ($settings as $k => $v) {
      $field->setThirdPartySetting('file_defaults_by_type', $k, $v)->save();
    }

    $entity_form_display = EntityFormDisplay::collectRenderDisplay($entity, 'default');
    $entity_form_display->setComponent('test_file_field', [
      'region' => 'content',
      'type' => 'default_file_choice',
    ]);
    $entity_form_display->save();

    $entity = EntityTest::create();
    $form_object = $this->container->get('entity_type.manager')->getFormObject('entity_test', 'default');
    $form_object->setEntity($entity);
    $form_state = new FormState();

    return $this->container->get('form_builder')->buildForm($form_object, $form_state);
  }

}
